Rotation = LibStub("AceAddon-3.0"):NewAddon("Rotation", "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0")
Rotation:SetDefaultModuleState(false)
Rotation:SetDefaultModuleLibraries("AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Rotation")

local defaults = {
	profile = {
		modules = {
			primary = nil,
			secondary = nil,
			primaryEnabled = false,
			secondaryEnabled = false,
		},
		display = {
			position = {
				left = 0,
				top = 0
			},
			size = {
				width = 36,
				height = 36
			},
			locked = false,
			border = 1,
			updateFrequency = 1/20 -- 20 times per second
		},
	},
}

function Rotation:InitialiseVariables()
	self.loadableAddons = {}
	self.classAddons = {}
	self.classModules = {}
	self.options = {}
	self.db = {}
	self.availableModulesSelect = {}
	self.activeModule = nil
end

function Rotation:LoadDatabase()
	self.db = LibStub("AceDB-3.0"):New("RotationDB", defaults)
end

function Rotation:OnInitialize()
	-- Code that you want to run when the addon is first loaded goes here.
	self:InitialiseVariables()
	self:LoadDatabase()
end

function Rotation:OnEnable()
    -- Called when the addon is enabled	
	self:GetLoadableAddons()
	self:GetClassAddons()
	self:GetClassModules()
	self:BuildOptions()
	self:BuildFrames()
	self:SelectActiveModule()
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED", "OnActiveTalentGroupChanged")
	
	self:Print(L["Enabled"])
end

function Rotation:OnDisable()
    -- Called when the addon is disabled
	Rotation:SelectActiveModule()
	self:Print(L["Disabled"])
end

function Rotation:OnUpdate(elapsed) 
	if self:IsEnabled() then
		self.sinceLastUpdate = self.sinceLastUpdate + elapsed
		if self.sinceLastUpdate >= self.db.profile.display.updateFrequency then
			self:ExecuteUpdate(self.sinceLastUpdate)
			self.sinceLastUpdate = 0
		end
	end
end

function Rotation:ExecuteUpdate(elapsed)
	if self.activeModule then
		if self.activeModule.OnUpdate then
			self.activeModule:OnUpdate(elapsed)
		end
	end
end

-- Start: Load and activate modules
function Rotation:EnableActiveModule()
	if self.Frame then
		-- Hide frame
		self.Frame:Show()
		-- Subscribe to events
		self.Frame:SetScript("OnUpdate", function(frame, elapsed) Rotation:OnUpdate(elapsed) end)
	end
	self.activeModule:Enable()
	
	self.sinceLastUpdate = 0
	self:ExecuteUpdate(0)
end

function Rotation:DisableActiveModule()
	self.activeModule:Disable()
end

function Rotation:NoActiveModule()
	if self.Frame then
		-- Hide frame
		self.Frame:Hide()
		-- Unsubscribe from events
		self.Frame:SetScript("OnUpdate", nil)
	end
end

function Rotation:GetModuleForActiveTalentGroup()
	local activeTalentGroup = GetActiveTalentGroup()
	if activeTalentGroup < 2 then
		if self.db.profile.modules.primaryEnabled then
			return self.db.profile.modules.primary
		end
	elseif activeTalentGroup == 2 then
		if self.db.profile.modules.secondaryEnabled then
			return self.db.profile.modules.secondary
		end
	end
	return nil
end

function Rotation:OnActiveTalentGroupChanged()
	if self:IsEnabled() then
		self:SelectActiveModule()
	end
end

function Rotation:SelectActiveModule()
	-- Disable currently active module
	if self.activeModule then self:DisableActiveModule() end
	-- Deselect current modile
	self.activeModule = nil
	-- Activate new module
	if self:IsEnabled() then
		local moduleToken = self:GetModuleForActiveTalentGroup()
		if moduleToken then
			if self.classModules[moduleToken] then
				self.activeModule = self.classModules[moduleToken]
				self:EnableActiveModule()
				return
			end
		end	
	end
	self:NoActiveModule()
end

function Rotation:GetLoadableAddons()
	local num = GetNumAddOns()
	local classes = {}
	FillLocalizedClassList(classes, (UnitSex("player") == 3))
	
	self.loadableAddons = {}
	
	for i = 1,num do
		local name, title, notes, enabled, loadable, reason, security = GetAddOnInfo(i)
		
		if name and enabled and loadable then
			local rotationName = GetAddOnMetadata(i, "X-Rotation-Name")
			local rotationClass = GetAddOnMetadata(i, "X-Rotation-Class")
			if rotationName and rotationClass then
				rotationClass = string.upper(rotationClass)
				if classes[rotationClass] then
					local tokenName = string.upper(rotationClass) .. "/" .. string.upper(rotationName)
				
					self.loadableAddons[tokenName] = {
					["name"] = rotationName,
					["displayName"] = rotationName,
					["tokenName"] = tokenName,	
					["addonName"] = name,
					["class"] = rotationClass,
					["tokenClass" ] = rotationClass,
					["localClass"] = classes[rotationClass],
					["index"] = i
					}					
				end
			end
		end
	end
end

function Rotation:GetClassAddons()
	local localClass, tokenClass = UnitClass("player")
	self.classAddons = {}
	for tokenName,addon in pairs(self.loadableAddons) do
		if addon.tokenClass == tokenClass then
			if IsAddOnLoaded(addon.index) or (IsAddOnLoadOnDemand(addon.index) and LoadAddOn(addon.index)) then
				self.classAddons[tokenName] = addon
			end
		end
	end
end

function Rotation:GetClassModules()
	local localClass, tokenClass = UnitClass("player")
	self.classModules = {}
	self.availableModulesSelect = {}
	for name, module in self:IterateModules() do
		if module:GetClassToken() == tokenClass then
			self.classModules[module:GetTokenName()] = module
			self.availableModulesSelect[module:GetTokenName()] = module:GetDisplayName()
		end
	end
end
-- End: Load and activate modules

-- Start: Options
function Rotation:BuildOptions()
	self.options = {
		type = "group",
		args = {
			modules = {
				name = L["Module Selection"],
				type = "group",
				handler = Rotation,
				order = 1,
				args = {
					primarytoggle = {
						order = 1,
						name = L["Enabled"],
						desc = L["Use a rotation module for your primary specialisation"],
						type = "toggle",
						get = "GetPrimaryEnabled",
						set = "SetPrimaryEnabled"
					},
					primary = {
						order = 2,
						name = L["Primary"],
						desc = L["Rotation module to use for your primary specialisation"],
						type = "select",
						style = "dropdown",
						values = self.availableModulesSelect,
						get = "GetPrimaryModule",
						set = "SetPrimaryModule",
						validate = "ValidateModule"
					},
					secondarytoggle = {
						order = 3,
						name = L["Enabled"],
						desc = L["Use a rotation module for your secondary specialisation"],
						type = "toggle",
						get = "GetSecondaryEnabled",
						set = "SetSecondaryEnabled"
					},
					secondary = {
						order = 4,
						name = L["Secondary"],
						desc = L["Rotation module to use for your secondary specialisation"],
						type = "select",
						style = "dropdown",
						values = self.availableModulesSelect,
						get = "GetSecondaryModule",
						set = "SetSecondaryModule",
						validate = "ValidateModule"	
					},
				},
			},
			display = {
				name = L["Display"],
				type = "group",
				handler = Rotation,
				order = 2,
				args = {
					lock = {
						order = 1,
						name = L["Lock Frame"],
						desc = L["Lock the frame so you do not accidentally move it"],
						type = "toggle",	
						get = "GetLocked",
						set = "SetLocked"
					},
					position = {
						order = 2,
						name = L["Position"],
						type = "group",
						inline = true,
						args = {
							left = {
								order = 1,
								name = L["Left"],
								type = "input",
								get = "GetPositionLeft",
								set = "SetPositionLeft",
								validate = "ValidatePosition"
							},
							top = {
								order = 2,
								name = L["Top"],
								type = "input",
								get = "GetPositionTop",
								set = "SetPositionTop",
								validate = "ValidatePosition"
							},
						},
					},
					size = {
						order = 3,
						name = L["Size"],
						type = "group",
						inline = true,
						args = {
							width = {
								order = 1,
								name = L["Width"],
								type = "input",
								get = "GetSizeWidth",
								set = "SetSizeWidth",
								validate = "ValidateSize"
							},
							height = {
								order = 2,
								name = L["Height"],
								type = "input",
								get = "GetSizeHeight",
								set = "SetSizeHeight",
								validate = "ValidateSize"
							},
						},
					},
				}
			},
			profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db),
		}
	}
	
	if GetNumTalentGroups() < 2 then
		self.options.args.modules.args.secondary.disabled = true
		self.options.args.modules.args.secondarytoggle.disabled = true
	end
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable(self:GetName(), self.options)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions(self:GetName(), self:GetName(), nil, "modules")
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions(self:GetName(), self.options.args.display.name, self:GetName(), "display")
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions(self:GetName(), self.options.args.profiles.name, self:GetName(), "profiles")
end

---- Start: Option Getters/Setters/Validators
function Rotation:GetPrimaryModule(info)
	return self.db.profile.modules.primary
end

function Rotation:SetPrimaryModule(info, val)
	self.db.profile.modules.primary = val
	Rotation:SelectActiveModule()
end

function Rotation:GetSecondaryModule(info)
	return self.db.profile.modules.secondary
end

function Rotation:SetSecondaryModule(info, val)
	self.db.profile.modules.secondary = val
	Rotation:SelectActiveModule()
end

function Rotation:ValidateModule(info, val)
	if not self.availableModulesSelect[val] then
		return L["Invalid module"]
	end
	return true
end

function Rotation:GetPrimaryEnabled(info)
	return self.db.profile.modules.primaryEnabled
end

function Rotation:SetPrimaryEnabled(info, val)
	self.db.profile.modules.primaryEnabled = val
	self.options.args.modules.args.primary.disabled = not val
	Rotation:SelectActiveModule()
end

function Rotation:GetSecondaryEnabled(info)
	return self.db.profile.modules.secondaryEnabled
end

function Rotation:SetSecondaryEnabled(info, val)
	self.db.profile.modules.secondaryEnabled = val
	self.options.args.modules.args.secondary.disabled = not val
	Rotation:SelectActiveModule()
end

function Rotation:GetLocked(info)
	return self.db.profile.display.locked
end

function Rotation:SetLocked(info, val)
	self.db.profile.display.locked = val
	Rotation:OnLockedChanged()
end

function Rotation:GetPositionLeft(info)
	return ""..self.db.profile.display.position.left
end

function Rotation:GetPositionTop(info)
	return ""..self.db.profile.display.position.top
end

function Rotation:GetSizeWidth(info)
	return ""..self.db.profile.display.size.width
end

function Rotation:GetSizeHeight(info)
	return ""..self.db.profile.display.size.height
end

function Rotation:SetPositionLeft(info,val)
	self.db.profile.display.position.left = tonumber(val)
	self:OnPositionChange()	
end

function Rotation:SetPositionTop(info,val)
	self.db.profile.display.position.top = tonumber(val)
	self:OnPositionChange()
end

function Rotation:SetSizeWidth(info,val)
	self.db.profile.display.size.width = tonumber(val)
	self:OnSizeChange()
end

function Rotation:SetSizeHeight(info,val)
	self.db.profile.display.size.height = tonumber(val)
	self:OnSizeChange()
end

function Rotation:ValidatePosition(info, val)
	if tonumber(val) < 0 then
		return L["Can not be less than"] .. "0"
	end
	return true
end

function Rotation:ValidateSize(info, val)
	if tonumber(val) < 10 then
		return L["Can not be less than"] .. "10"
	end
	return true
end
---- End: Option Getters/Setters/Validators
-- End: Options

-- Start: Module creation with custom prototype
function Rotation:CreateModule(name, class)
	local prototype = {
		moduleClass = class,
		moduleName = name,
		moduleToken = string.upper(class) .. "/" .. string.upper(name),
		OnEnable = function(self)
			self:Print("Enabled")
		end,
		OnDisable = function(self)
			self:Print("Disabled")
		end,
		GetClassToken = function(self)
			return string.upper(self.moduleClass)
		end,
		GetDisplayName = function(self)
			return self.moduleName
		end,
		GetTokenName = function(self)
			return self.moduleToken
		end	
	}
	return self:NewModule(name, prototype)
end
-- End: Module creation with custom prototype

function Rotation:OnLockedChanged()
	if self.Frame then
		self.Frame:SetMovable(not self.db.profile.display.locked)
		self.Frame:SetResizable(not self.db.profile.display.locked)
		self.Frame:EnableMouse(not self.db.profile.display.locked)
		if not self.db.profile.display.locked then			
			self.Frame:SetScript("OnDragStart", function() Rotation:OnDragStart() end)
			self.Frame:SetScript("OnDragStop", function() Rotation:OnDragStop() end)
			self.Frame:RegisterForDrag("LeftButton");
		else
			self.Frame:SetScript("OnDragStart", nil)
			self.Frame:SetScript("OnDragStop", nil)
			self.Frame:RegisterForDrag(nil);
		end
	end
end

function Rotation:OnDragStart()
	local _, _, _, x, y = self.Frame:GetPoint()
	self.dragStartFrameX = x
	self.dragStartFrameY = y
	
	self.Frame:StartMoving()
	
	_, _, _, x, y = self.Frame:GetPoint()
	self.dragStartScreenX = x
	self.dragStartScreenY = y
end

function Rotation:OnDragStop()
	local _, _, _, x, y = self.Frame:GetPoint()
	self.dragStopScreenX = x
	self.dragStopScreenY = y
	
	self.Frame:StopMovingOrSizing()
	
	local dx = self.dragStopScreenX - self.dragStartScreenX 
	local dy = self.dragStopScreenY - self.dragStartScreenY
	
	self.db.profile.display.position.left = self.dragStartFrameX + dx
	self.db.profile.display.position.top = -(self.dragStartFrameY + dy)
	
	Rotation:OnPositionChange()
end

function Rotation:OnPositionChange()
	-- Update frame
	self.Frame:ClearAllPoints()
	self.Frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", self.db.profile.display.position.left, -self.db.profile.display.position.top)
end

function Rotation:OnSizeChange()
	-- Update frame
	self.Frame:SetWidth(self.db.profile.display.size.width)
	self.Frame:SetHeight(self.db.profile.display.size.height)
end

function Rotation:BuildFrames()
	if not self.Frame then self.Frame = CreateFrame("Frame", "Rotation_Frame", UIParent) end
	if self.Frame then
		self.Frame:Hide()
		self.Frame:SetFrameLevel(1)
		self.Frame:SetFrameStrata("HIGH")
		self.Frame:SetWidth(self.db.profile.display.size.width)
		self.Frame:SetHeight(self.db.profile.display.size.height)
		self.Frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", self.db.profile.display.position.left, -self.db.profile.display.position.top)
		self:OnLockedChanged()
		if not self.FrameBackground then self.FrameBackground = self.Frame:CreateTexture()	end
		if self.FrameBackground then
			self.FrameBackground:SetDrawLayer("BACKGROUND")
			self.FrameBackground:SetAllPoints(self.Frame)
			self.FrameBackground:SetTexture(0.0, 0.0, 0.0, 1.0)
		end
		if not self.FrameIcon then self.FrameIcon = self.Frame:CreateTexture()	end
		if self.FrameIcon then
			self.FrameIcon:SetDrawLayer("ARTWORK")
			self.FrameIcon:SetPoint("TOPLEFT", self.Frame, "TOPLEFT", self.db.profile.display.border, -self.db.profile.display.border)
			self.FrameIcon:SetPoint("BOTTOMRIGHT", self.Frame, "BOTTOMRIGHT", -self.db.profile.display.border, self.db.profile.display.border)
			self.FrameIcon:SetTexture("Interface\\AddOns\\Rotation\\Media\\placeholder")
		end
		if not self.FrameCooldown then self.FrameCooldown = CreateFrame("Cooldown", nil, self.Frame) end
		if self.FrameCooldown then
			self.FrameCooldown:SetAllPoints()
			self.FrameCooldown:SetDrawEdge()
		end	
	end
end

function Rotation:SetSpell(spell, target)
	if spell then
		target = target or "target"
		local name, rank, icon, powerCost, isFunnel, powerType, castingTime,  minRange, maxRange = GetSpellInfo(spell)
		if name then
			local isUsable, notEnoughMana = IsUsableSpell(name)
			self.FrameIcon:SetTexture(icon)
			
			-- Range/Mana check
			if IsSpellInRange(spell, target) == 1 then
				if notEnoughMana == 1 then
					-- not enough mana, blue border
					self.FrameBackground:SetTexture(0.0, 0.0, 1.0, 1.0)
				elseif isUsable == 1 then
					-- useable
					self.FrameBackground:SetTexture(0.0, 1.0, 0.0, 1.0)
				else
					-- not usable
					self.FrameBackground:SetTexture(0.0, 0.0, 0.0, 1.0)
				end
			else
				-- Out of range, red border
				self.FrameBackground:SetTexture(1.0, 0.0, 0.0, 1.0)
			end
			
			-- Time
			local start, duration, enable = GetSpellCooldown(spell)
			if start > 0 then
				self.FrameCooldown:Show()
				self.FrameCooldown:SetCooldown(start, duration)
			else
				self.FrameCooldown:Hide()
			end
		else
			-- Error, red border
			self.FrameBackground:SetTexture(1.0, 0.0, 0.0, 1.0)
			self.FrameIcon:SetTexture("Interface\\AddOns\\Rotation\\Media\\placeholder")
			self.FrameCooldown:Hide()
		end
	else
		-- No spell, black border
		self.FrameBackground:SetTexture(0.0, 0.0, 0.0, 1.0)
		self.FrameIcon:SetTexture("Interface\\AddOns\\Rotation\\Media\\placeholder")
		self.FrameCooldown:Hide()
	end	
end

